/*
 *  File: clp_cfg.c 		*PZ*		28 Nov 1998
 */

/*
 * This is the module for customizing CLP support in a CSP-II
 *	runtime system linked by the customer.
 *
 * Up to 4 CLP solvers (or none at all) can be included
 *	in the runtime. The configuration must be described in a
 *	user-prepared table which is generated by customizing and
 *	compiling this module. Even the same solver can be configured
 *	more than once, under different names (in this case the solver
 *	must be reentrant, and maintain a usage count for package
 *	initialization/termination, and must not use static cash).
 *
 * The source of the module is shipped with a default configuration
 *	consisting of only the simple linear solver provided by ML.
 *
 * All customization is to be confined to this module; the included
 *	'clp_cfg.h' file must not be changed.
 *
 * The macro definitions used in the example serve also as comments;
 *	the user can directly enter the desired values into the
 *	initialization part of the data structure (array).
 *
 * The comments surrounding the individual definitions should be
 *	sufficient aid for the user in preparing her/his own
 *	configuration properly. Also refer to the comments
 *	in 'clp_cfg.h'.
 *
 * The compilation can be simplified for the default cases (the source
 *		module need not be changed for these):
 *	(1) If no macro definition is specified, then an empty
 *	    configuration table is genereted (N_SOLVERS = 0)
 *	(2) If only  '-DCONSTRAINT'  is specified,
 *	    then the generated configuration table will consist of 
 *	    one item - the default (which is originally the ML solver).
 *
 *   Otherwise  '-DN_SOLVERS=<number> can be specified for the general
 *	case.
 */

/*
 * The capability of individual solvers described by separate entry field
 *		in the congiguration table is the following:
 *
 *	- Cooperativness, which can be
 *	   -- monopolistic
 *		cannot run together with another solver instance
 *		in the same thread. An instance of this solver type
 *		would not be launched if the thread already has an
 *		instance of another solver running; and if
 *		launched as first, does not allow other instances
 *		to be launched. (Probable reason: resource-intensive
 *		or needs some special resource for exclusive use).
 *	   -- xenophobic (cannot cope with constrained variables 
 *		belonging to other solver - the core system must
 *		pre-filter expressions passed to this solver and
 *		reject those containing 'aliens').
 *	   -- tolerant (can live together with other solvers
 *		in a thread - recognizes and rejects 'alien'
 *		constrained variables owned by other solvers.
 *	   -- cooperative (not supported yet - probably needs a
 *		compatibility matrix - can cooperate in forming
 *		clp_value, clp_max, clp_min; in unification
 *		might or might not use the restrictions imposed by
 *		other solvers.
 */


	/************************************************
	*  Either specify N_SOLVER here (uncomment!),   *
	*	or supply the desired value externally, *
	*	or accept the default value of 0.	*
	************************************************/
/***
#define	N_SOLVERS	0
***/
	/* ******************************************* */


#include <stdlib.h>

#include <clp_cfg.h>



	/****************************************
	*					*
	*  Here come the definitions that	*
	*	are subject to customization	*
	*					*
	****************************************/




#if N_SOLVERS >= 1	/* ============================== */

	/*
	 * Parameters for Solver-0 (the default solver):
	 */

#define S0_CALL_INTERFACE_TABLE		XxpMLLinearSolverIf
#define	S0_PROGRAMMATIC_NAME		"ml_linear_solver"

extern	const	ClpIfTable	S0_CALL_INTERFACE_TABLE;

#endif /* >= 1 */	/* ============================== */


#if N_SOLVERS >= 2	/* ============================== */

	/*
	 * Parameters for Solver-1):
	 */

#define S1_CALL_INTERFACE_TABLE		<external symbol>
#define	S1_PROGRAMMATIC_NAME		"<text for CSP atom>"

extern	const	ClpIfTable	S1_CALL_INTERFACE_TABLE;

#endif /* >= 2 */	/* ============================== */

		/*  ...  */



		/*******************************************
		*  Declare and fill the descriptor table:  *
		*******************************************/


	ClpSolverDescr	XxpSolverTable	= {

	  N_SOLVERS,

#if N_SOLVERS >= 1
	  {
		{ &S0_CALL_INTERFACE_TABLE,
			S0_PROGRAMMATIC_NAME,	ClpCoopTolerant,	},

#if N_SOLVERS >= 2
		{ &S1_CALL_INTERFACE_TABLE,
			S1_PROGRAMMATIC_NAME,	ClpCoop???,		},
		
#if N_SOLVERS >= 3
		{ &S2_CALL_INTERFACE_TABLE,
			S2_PROGRAMMATIC_NAME,	ClpCoop???,		},
		
#if N_SOLVERS >= 4
		{ &S3_CALL_INTERFACE_TABLE,
			S3_PROGRAMMATIC_NAME,	ClpCoop???,		},

		/***************************
		*  (end of customization)  *
		***************************/

#endif /* >= 4 */
#endif /* >= 3 */
#endif /* >= 2 */
	  }
#endif /* >= 1 */
	};
/* -------------------------------------------- */
/* -------------------------------------------- */
/**************************************
 NOTES:
   1. The unfilled slots are initialized to zeros by the system.
   2. The filled part must be dense, at the beginning of the array
	('holes' are not allowed).
   3. All 'solver_name' (S<i>_PROGRAMMATIC_NAME) entries must be
	unique.
   4. The solver occupying slot-0 is the default solver; it is selected
	initially in each CSP process, and also by the call of
	'select_clp_solver/0' predicate.
   5. The index of the entry within the array serves as SolverId
	for the solver named in the entry.
	This SolverId is used in the external representation of
	the constrained variables (e.g. when written out by one of the
	'write' or 'format' predicates and in system messages).
	The Id of the default solver (0) is suppressed in the
	synthesized variable names (the same way as ProcessId == 0
	is suppressed).
   6. N_SOLVERS can be supplied externally (from the makefile, or
	as a direct command line argument) for the C or C++ compiler
	used. By using this feature it is possible to select the
	initial part of any size of a prepared table without changing
	the source module itself.
   7. A 'really monopolistic' solver kind can be defined only indirectly,
	as a configuration with only that solver, with no multi-threading
	support.
   8. An important characteristic of the solver, which earlier also had 
	a separate field in the configuration table is:
		- Single- or multi-threaded execution.
		   If the solver can have separate data area for several
		   problems being solved in quasi-paralell - each problem
		   belonging to a different CSP process, then it is qualified
		   as multithreding solver. Otherwise it is single-threaded.
	Now this feature is identified by setting the 'swap' entry in the 
	interface entry points table to NULL or non-NULL.
**************************************/
