/*
 * file: net_hello.pro		ML Ltd		04Nov97
 *	- Simple test for demonstrating that networking works at all
 */

:- module(net_hello, main_goal/1).


/*
 * Motto: "In my dream I was two cats and played with each other" (Karinthy Fr.)
 *
 * Usage:
 *   Two instances of the program will exchange a few messages through the
 *   network. The roles are controlled by the first command line argument, which
 *   should be 'active' for the one that initiates conversation, and
 *   'reactive' for the other, that will react.
 *
 *   If the second (optional) argument is specified as 'alerts', then the program
 *      instance writes out all alerts. Each alert appears as a framed message
 *      for better visual separation.
 *
 *   In this simple example both instances will run on the local host.
 *
 *
 * The simplest way to run the program is to issue a shell command like
 *	the following one:
 *
 *	    csprolog -cspprog net_hello active & csprolog -cspprog net_hello reactive alerts
 *
 *	(in this case, however, the output from both will be intermixed on
 *	the common console output)
 *
 *  Beside alerts, the following conversation should be seen if the outputs
 *	are coalesced:
 *
 * 		A: hello (being sent)
 *						R: hello (received)
 *						R: welcome (being sent)
 *		A: welcome (received)
 *		A: bye (being sent)
 *						R: bye (received)
 */

#define aPORT	6872		% ip_port for the active side
#define rPORT	6913		% ip_port for the reactive side


main_goal(L) :-
  L = [ROLE | OPT], member([active, reactive], ROLE),
			member([[alerts], []], OPT) -> talk(ROLE, OPT);
  format('   *** Invalid argument list: ~q ***~2n  Usage:  net_hello { active | reactive } [ alerts ]~n', [L]).


talk(ROLE, OPT) :-
  member([ [active, 'A', active_port, reactive_port, aPORT, rPORT, graceful],
	   [reactive, 'R', reactive_port, active_port, rPORT, aPORT, graceful] ],
		 [ROLE, PFX, PORT_NAME, PEER_PORT_NAME, OUR_TCP_PORT, PEER_TCP_PORT, CLOSE_MODE]),
    new_rt(alert_proc, alert_goal(PFX), alert_init(OPT), [netalert, control], period(500)),
    start_processes,
    community_init(slave, ROLE, [], netalert, 0),
    port_create(PORT_NAME, chan_in, off),
    community_activate([[self, name(myself), ip_port(OUR_TCP_PORT)],
			[name(PEER), ip_port(PEER_TCP_PORT), max_retries(24), retry_delay(500)]]),
    connect_to_port(line, PEER, PEER_PORT_NAME, chan_out),
    open_channel_for_send(chan_out),
    open_channel_for_receive(chan_in),
    !,
    do_the_talking(ROLE, PFX),
    close_channel(chan_out),
    close_channel(chan_in),
    disconnect(line),
    port_close(PORT_NAME),
    generate_event(control, shut(CLOSE_MODE)).

do_the_talking(active, PFX) :- !,
  Indent = 0,
    report_and_send(hello, PFX, Indent),
    receive_and_report(welcome, PFX, Indent),
    report_and_send(bye, PFX, Indent).
 
do_the_talking(reactive, PFX) :-
  Indent = 25,
    receive_and_report(hello, PFX, Indent),
    report_and_send(welcome, PFX, Indent),
    receive_and_report(bye, PFX, Indent).

report_and_send(Msg, Prefix, Indent) :-
  format('~N  ~*|~a: ~q (being sent)~n', [Indent, Prefix, Msg]),
    send(chan_out, Msg).

receive_and_report(Msg, Prefix, Indent) :-
  receive(chan_in, X),
    format('~N  ~*+~a: ~q (received)~n', [Indent, Prefix, X]),
    (X == Msg -> true ; throw('Unexpected message instead of awaited'(X, Msg)) ).


:-meta_predicate(member(+,?)).
member([M|_], M).
member([_|T], M) :- member(T, M).

work(0) :- !.
work(N) :-
  deschedule_process,
    N1 is N - 1,
    work(N1).



	/***************************************
	*  Alert display & sequencing control  *
	***************************************/


% --------------------------------------------------------
%  Telling apart 'past' from 'future' using cyclic clocks
% --------------------------------------------------------

#define cYCLE		8640000		/* 24 * 60 * 60 * 100 */
#define hALF_CYCLE	4320000		/* 12 * 60 * 60 * 100 */

:-meta_predicate(expiration_time(+,-)).
:-meta_predicate(still_pending(+)).

#if 0
expiration_time(Delay, When) :-
      wall_clock_time(Now),
	When is (Now + Delay) mod cYCLE.

still_pending(ExpTime) :-
      wall_clock_time(Now),
	( Now =< ExpTime, Now + hALF_CYCLE > ExpTime -> true;
	  Now > ExpTime + hALF_CYCLE
	).
#endif

%  This should work if Delay is always less than hALF_CYCLE and the test
%    is performed not later than a hALF_CYCLE after expiration.
%
%  A simplified version (less evident):

expiration_time(Delay, When) :-
      wall_clock_time(Now),
	When is Now + Delay.

still_pending(ExpTime) :-
      wall_clock_time(Now),
	Now =< ExpTime,
	Now + hALF_CYCLE > ExpTime.

timeout_over(ExpTime) :-
      wall_clock_time(Now),
	( Now > ExpTime -> true; Now + hALF_CYCLE =< ExpTime ).

% ------------ (end of fast-future recognition) ----------


todo(Task, Delay) :-
      expiration_time(Delay, T),
	set_value(next_task, [Task, T]).

do_abort :-
      todo(close, 1000),
	community_shut(force).

opt_abort(Endtime, _, _) :-
      still_pending(Endtime), !.
opt_abort(_, PFX, Marker) :-
      format('   *** ~w: ~w has stuck ***~n', [PFX, Marker]),
	do_abort.


alert_init(L) :-
	( L == [alerts] -> set_value(switch, show) ;
	  L == [] -> true ;
		format('  *** Unexpected option(s): ''~q'', ignored ***~n', [L])
	),
	  todo(activate, 2000).


alert_goal(PFX) :-		%  alert event
	get_event(netalert, X) ->
		( test_value(switch) ->
			 format('~n~78c~n|  ~a: Alert: ~q  ~77||~n~78c~n', [0'-, PFX, X, 0'-]) ;
		  true
		),
		( X \= [community_shut_succeeded | _] -> true ; halt(0) ).

alert_goal(PFX) :-		%  control commannd
	get_event(control, Cmd) ->
		( Cmd = shut(Mode) -> todo(close, 1000), community_shut(Mode) ;
			format('   *** ~w: Unhandled control command \'~w\' ***~n', [PFX]),
			  halt(1)
		).

alert_goal(PFX) :-		% timer event  - sequencing control
      get_value(next_task, [Task, Endtime]),
	(
	  Task = activate ->
	      community_current_attribute(_, status, St),
	      ( St == closing -> todo(close, 1000) ;
	        St == normal  -> todo(connect, 30000) ;		% 5 minutes
		St == zombie  -> 
			format('   *** ~w: Community initialization failed ***~n', [PFX]),
			  do_abort;
		St == initializing -> opt_abort(Endtime, PFX, 'Community initialization');
			format('   *** ~w: Unexpected community status: ~w ***~n', [PFX, St]),
			  halt(1)
	      );

	  Task = connect ->
	      connection_current_attribute(line, status, St),
	      ( St == normal -> todo(wait, 6000) ;
		St == zombie -> 
			format('   *** ~w: Cannot connect to partner ***~n', [PFX]),
			  do_abort;
		    opt_abort(Endtime, PFX, 'Connection to partner')
	      );

	  Task = wait -> opt_abort(Endtime, PFX, 'Conversation');	% initializing, closing, <other?>

	  Task = close ->
	        ( community_current_attribute(_, status, St, conditional) ->
			( member([closing, zombie], St) -> true;
			  format('   *** ~w: Unexpected community status for closing ~w ***~n', [PFX, St]),
			    halt(1)
			),
			( still_pending(Endtime) -> true; 
			      format('   *** ~w: Community closing stuck ***~n', [PFX]),
				halt(1)
			);
		    format('   * ~w: Ended (by force) *~n', [PFX]),
		      halt(0)
		 ) ;

	  format('   *** ~w: Unhandled task from control schedule ~w ***~n', [PFX, Task]),
	    halt(1)
	).	
/* ---------------------------------------------- */
/* ---------------------------------------------- */
